#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

import logging

import venusian
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.interfaces import IRequest, IView
from pyramid.threadlocal import get_current_registry
from zope.component.interfaces import ComponentLookupError
from zope.contentprovider.interfaces import BeforeUpdateEvent
from zope.interface import Interface, classImplements, implementer
from zope.location.interfaces import ILocation

from pyams_template.template import get_view_template
from pyams_utils.request import check_request
from pyams_viewlet.interfaces import IViewlet, IViewletManager


logger = logging.getLogger('PyAMS (viewlet)')


@implementer(IViewletManager)
class ViewletManager(object):
    """The Viewlet Manager base

    A generic manager class which can be instantiated
    """

    permission = None
    template = None

    def __init__(self, context, request, view):
        self.__updated = False
        self.__parent__ = view
        self.context = context
        self.request = request

    def __getitem__(self, name):
        """See zope.interface.common.mapping.IReadMapping"""
        # Find the viewlet
        registry = get_current_registry()
        viewlet = registry.queryMultiAdapter((self.context, self.request, self.__parent__, self),
                                             IViewlet, name=name)

        # If the viewlet was not found, then raise a lookup error
        if viewlet is None:
            raise ComponentLookupError('No provider with name `%s` found.' % name)

        # If the viewlet cannot be accessed, then raise an
        # unauthorized error
        if viewlet.permission and not self.request.has_permission(viewlet.permission, context=self.context):
            raise HTTPUnauthorized('You are not authorized to access the provider called `%s`.' % name)

        # Return the viewlet.
        return viewlet

    def get(self, name, default=None):
        """See zope.interface.common.mapping.IReadMapping"""
        try:
            return self[name]
        except (ComponentLookupError, HTTPUnauthorized):
            return default

    def __contains__(self, name):
        """See zope.interface.common.mapping.IReadMapping"""
        return bool(self.get(name, False))

    def filter(self, viewlets):
        """Sort out all content providers

        ``viewlets`` is a list of tuples of the form (name, viewlet).
        """
        # Only return viewlets accessible to the principal
        request = self.request
        return [(name, viewlet) for name, viewlet in viewlets
                if (not viewlet.permission) or request.has_permission(viewlet.permission, context=self.context)]

    def sort(self, viewlets):
        """Sort the viewlets.

        ``viewlets`` is a list of tuples of the form (name, viewlet).
        """
        # By default, we are not sorting by viewlet name.
        return sorted(viewlets, key=lambda x: x[0])

    def update(self):
        """See zope.contentprovider.interfaces.IContentProvider"""
        # check permission
        if self.permission and not self.request.has_permission(self.permission, context=self.context):
            return
        # get the viewlets from now on
        self.viewlets = []
        append = self.viewlets.append
        for name, viewlet in self._get_viewlets():
            if ILocation.providedBy(viewlet):
                viewlet.__name__ = name
            append(viewlet)
        # and update them...
        self._update_viewlets()
        self.__updated = True

    def _get_viewlets(self):
        """Find all content providers for the region"""
        registry = self.request.registry
        viewlets = registry.getAdapters((self.context, self.request, self.__parent__, self),
                                        IViewlet)
        viewlets = self.filter(viewlets)
        viewlets = self.sort(viewlets)
        return viewlets

    def _update_viewlets(self):
        """Calls update on all viewlets and fires events"""
        registry = self.request.registry
        for viewlet in self.viewlets:
            registry.notify(BeforeUpdateEvent(viewlet, self.request))
            viewlet.update()

    def render(self):
        """See zope.contentprovider.interfaces.IContentProvider"""
        # Check for previous update
        if not (self.__updated and self.viewlets):
            return ''
        # Now render the view
        if self.template:
            return self.template(viewlets=self.viewlets)
        else:
            return '\n'.join([viewlet.render() for viewlet in self.viewlets])


def ViewletManagerFactory(name, interface, bases=(), cdict=None):
    """Viewlet manager factory"""

    attr_dict = {'__name__': name}
    attr_dict.update(cdict or {})

    if ViewletManager not in bases:
        # Make sure that we do not get a default viewlet manager mixin, if the
        # provided base is already a full viewlet manager implementation.
        if not (len(bases) == 1 and IViewletManager.implementedBy(bases[0])):
            bases = bases + (ViewletManager,)

    viewlet_manager_class = type('<ViewletManager providing %s>' % interface.getName(), bases, attr_dict)
    classImplements(viewlet_manager_class, interface)
    return viewlet_manager_class


def get_weight(item):
    """Get sort weight of a given viewlet"""
    name, viewlet = item
    try:
        return int(viewlet.weight)
    except (TypeError, AttributeError):
        return 0


def get_label(item, request=None):
    """Get sort label of a given viewlet"""
    name, viewlet = item
    try:
        if request is None:
            request = check_request()
        return request.localizer.translate(viewlet.label)
    except AttributeError:
        return '--'


def get_weight_and_label(item, request=None):
    """Get sort weight and label of a given viewlet"""
    return get_weight(item), get_label(item, request)


class WeightOrderedViewletManager(ViewletManager):
    """Weight ordered viewlet managers.

    Viewlets with the same weight are sorted by label
    """

    def sort(self, viewlets):
        return sorted(viewlets, key=lambda x: get_weight_and_label(x, request=self.request))


def is_available(viewlet):
    try:
        return ((not viewlet.permission) or
                viewlet.request.has_permission(viewlet.permission, context=viewlet.context)) and \
               viewlet.available
    except AttributeError:
        return True


class ConditionalViewletManager(WeightOrderedViewletManager):
    """Conditional weight ordered viewlet managers"""

    def filter(self, viewlets):
        """Sort out all viewlets which are explicit not available

        ``viewlets`` is a list of tuples of the form (name, viewlet).
        """
        return [(name, viewlet) for name, viewlet in viewlets if is_available(viewlet)]


class TemplateBasedViewletManager(object):
    """Template based viewlet manager mixin class"""

    template = get_view_template()


class viewletmanager_config(object):
    """Class or interface decorator used to declare a viewlet manager

    You can provide same arguments as in 'viewletManager' ZCML directive:
    @name = name of the viewlet; may be unique for a given viewlet manager
    @view = the view class or interface for which viewlet is displayed
    @for_ = the context class or interface for which viewlet is displayed
    @permission = name of a permission required to display the viewlet
    @layer = request interface required to display the viewlet
    @class_ = the class handling the viewlet manager; if the decorator is applied
      on an interface and if this argument is not provided, the viewlet manager
      will be handled by a default ViewletManager class
    @provides = an interface the viewlet manager provides; if the decorator is
      applied on an Interface, this will be the decorated interface; if the
      decorated is applied on a class and if this argument is not specified,
      the manager will provide IViewletManager interface.
    """

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        if not settings.get('name'):
            raise ConfigurationError("You must provide a name for a ViewletManager")
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, ob):
            cdict = {'__name__': settings.get('name')}
            if 'permission' in settings:
                cdict['permission'] = settings.get('permission')

            if issubclass(ob, Interface):
                class_ = settings.get('class_', ViewletManager)
                provides = ob
            else:
                class_ = ob
                provides = settings.get('provides', IViewletManager)
            new_class = ViewletManagerFactory(settings.get('name'), provides, (class_,), cdict)

            logger.debug("Registering viewlet manager {0} ({1})".format(settings.get('name'),
                                                                        str(new_class)))
            config = context.config.with_package(info.module)
            config.registry.registerAdapter(new_class,
                                            (settings.get('context', Interface),
                                             settings.get('layer', IRequest),
                                             settings.get('view', IView)),
                                            provides, settings.get('name'))

        info = self.venusian.attach(wrapped, callback, category='pyams_viewlet')

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped
